from __future__ import print_function

import datetime
import requests
import arrow
import pytz

from .core import Service, NoData
from .crypto_data import crypto_data

quandl_exchange_btc_to_fiat = {
    'ARS': 'localbtc',
    'AUD': 'weex',
    'BRL': 'bbm',
    'CAD': 'virtex',
    'CHF': 'localbtc',
    'CLP': 'th',
    'CNY': 'anxhk',
    'CZK': 'bitcash',
    'DKK': 'localbtc',
    'EUR': 'btce',
    'GAU': 'bcmBM',
    'GBP': 'britcoin',
    'HKD': 'anxhk',
    'HUF': 'ruxum',
    'ILS': 'bit2c',
    'INR': 'localbtc',
    'JPY': 'btcex',
    'MXN': 'localbtc',
    'NOK': 'just',
    'NZD': 'bitnz',
    'PLN': 'bitcurex',
    'RUB': 'ruxum',
    'RUR': 'btce',
    'SEK': 'ruxum',
    'SGD': 'ruxum',
    'SLL': 'virwox',
    'THB': 'ruxum',
    'UAH': 'ruxum',
    'WMR': 'btcex',
    'WMZ': 'btcex',
    'XRP': 'ripple',
    'YAD': 'btcex',
    'ZAR': 'bitx',
}

class Quandl(Service):
    def get_historical_price(self, crypto, fiat, at_time):
        """
        Using the quandl.com API, get the historical price (by day).
        The CRYPTOCHART source claims to be from multiple exchange sources
        for price (they say best exchange is most volume).
        """
        # represents the 'width' of the quandl data returned (one day)
        # if quandl ever supports data hourly or something, this can be changed
        interval = datetime.timedelta(hours=48)
        crypto = crypto.lower()
        fiat = fiat.lower()

        at_time = arrow.get(at_time).datetime

        data = crypto_data[crypto]
        name, date_created = data['name'], data['genesis_date']

        if date_created.replace(tzinfo=pytz.utc) > at_time:
            raise Exception("%s (%s) did not exist on %s" % (name, crypto, at_time))

        if crypto == 'btc':
            # Bitcoin to fiat
            if fiat == 'usd':
                if at_time < datetime.datetime(2013, 2, 1, tzinfo=pytz.utc):
                    exchange = 'MtGox'
                else:
                    exchange = "Bitstamp"
            else:
                exchange = quandl_exchange_btc_to_fiat[fiat.upper()]

            source = "BITCOIN/%s%s" % (exchange.upper(), fiat.upper())
            price_index = 1
        else:
            # some altcoin to bitcoin
            if fiat != 'btc':
                raise Exception("Altcoins are only available via BTC base fiat")
            if crypto == 'ltc':
                source, price_index = ['BTCE/BTCLTC', 4]
            else:
                source, price_index = ['CRYPTOCHART/' + crypto.upper(), 1]

        url = "https://www.quandl.com/api/v1/datasets/%s.json" % source
        trim = "?trim_start={0:%Y-%m-%d}&trim_end={1:%Y-%m-%d}".format(
            at_time - interval, at_time + interval
        )

        response = self.get_url(url + trim).json()
        closest_distance = interval

        best_price = None
        for line in response['data']:
            date = line[0]
            price = line[price_index]
            tick_date = arrow.get(date).datetime
            distance = at_time - tick_date

            if distance.total_seconds() == 0:
                return price, source, tick_date

            if distance < closest_distance:
                closest_distance = distance
                best_price = price
                best_date = tick_date

        if not best_price:
            msg = "Quandl's data source is incomplete. Could not get best price for %s/%s on %s." % (
                crypto, fiat, at_time
            )
            raise NoData(msg)

        return best_price, source, best_date



from __future__ import print_function

from .core import AutoFallback, enforce_service_mode
from .historical_price import Quandl
from .crypto_data import crypto_data


def _get_optimal_services(crypto, type_of_service):
    try:
        # get best services from curated list
        return crypto_data[crypto]['services'][type_of_service]
    except KeyError:
        raise ValueError("Invalid cryptocurrency symbol: %s" % crypto)

def get_current_price(crypto, fiat, services=None, **modes):
    if not services:
        services = _get_optimal_services(crypto, 'current_price')

    return enforce_service_mode(
        services, CurrentPrice, {'crypto': crypto, 'fiat': fiat}, modes=modes
    )


def get_address_balance(crypto, address, services=None, **modes):
    if not services:
        services = _get_optimal_services(crypto, 'address_balance')

    return enforce_service_mode(
        services, AddressBalance, {'crypto': crypto, 'address': address}, modes=modes
    )


def get_historical_transactions(crypto, address, services=None, **modes):
    if not services:
        services = _get_optimal_services(crypto, 'historical_transactions')

    return enforce_service_mode(
        services, HistoricalTransactions, {'crypto': crypto, 'address': address}, modes=modes
    )


def get_unspent_outputs(crypto, address, services=None, **modes):
    if not services:
        services = _get_optimal_services(crypto, 'unspent_outputs')
    return enforce_service_mode(
        services, UnspentOutputs, {'crypto': crypto, 'address': address}, modes=modes
    )


def get_historical_price(crypto, fiat, date):
    """
    Only one service is defined fr geting historical price, so no fetching modes
    are needed.
    """
    return HistoricalPrice().get(crypto, fiat, date)


def push_tx(crypto, tx_hex, verbose=False, services=None, **modes):
    if not services:
        services = _get_optimal_services(crypto, 'push_tx')
    return enforce_service_mode(
        services, PushTx, {'crypto': crypto, 'tx_hex': tx_hex}, modes=modes
    )

def get_block(crypto, block_number='', block_hash='', latest=False, services=None, **modes):
    if not services:
        services = _get_optimal_services(crypto, 'get_block')
    kwargs = dict(crypto=crypto, block_number=block_number, block_hash=block_hash, latest=latest)
    return enforce_service_mode(
        services, GetBlock, kwargs, modes=modes
    )


def get_optimal_fee(crypto, tx_bytes, acceptable_block_delay, verbose=False):
    return OptimalFee(verbose=verbose).get(crypto, tx_bytes, acceptable_block_delay)


class OptimalFee(AutoFallback):

    def get(self, crypto, tx_bytes, acceptable_block_delay=0):
        crypto = crypto.lower()
        return self._try_each_service("get_optimal_fee", crypto, tx_bytes, acceptable_block_delay)

    def no_service_msg(self, crypto, tx_bytes, acceptable_block_delay):
        return "Could not get optimal fee for: %s" % crypto

class GetBlock(AutoFallback):

    def get(self, crypto, block_number='', block_hash='', latest=False):
        if sum([bool(block_number), bool(block_hash), bool(latest)]) != 1:
            raise ValueError("Only one of `block_hash`, `latest`, or `block_number` allowed.")
        return self._try_each_service(
            'get_block', crypto, block_number=block_number, block_hash=block_hash, latest=latest
        )

    def no_service_msg(self, crypto, block_number='', block_hash='', latest=False):
        return "Could not get %s block: %s%s%s" % (
            crypto, block_number, block_hash, 'latest' if latest else ''
        )

    @classmethod
    def strip_for_consensus(self, results):
        stripped = []
        for result in results:
            stripped.append(
                "[hash: %s, number: %s, size: %s]" % (
                    result['hash'], result['block_number'], result['size']
                )
            )
        return stripped

class HistoricalTransactions(AutoFallback):

    def get(self, crypto, address):
        return self._try_each_service('get_transactions', crypto, address)

    def no_service_msg(self, crypto, address):
        return "Could not get transactions for: %s" % crypto

    @classmethod
    def strip_for_consensus(cls, results):
        stripped = []
        for result in results:
            result.sort(key=lambda x: x['date'])
            stripped.append(
                ", ".join(
                    ["[id: %s, amount: %s]" % (x['txid'], x['amount']) for x in result]
                )
            )
        return stripped

class UnspentOutputs(AutoFallback):

    def get(self, crypto, address):
        return self._try_each_service('get_unspent_outputs', crypto=crypto, address=address)

    def no_service_msg(self, crypto, address):
        return "Could not get unspent outputs for: %s" % crypto

    @classmethod
    def strip_for_consensus(cls, results):
        stripped = []
        for result in results:
            result.sort(key=lambda x: x['output'])
            stripped.append(
                ", ".join(
                    ["[output: %s, value: %s]" % (x['output'], x['amount']) for x in result]
                )
            )
        return stripped


class CurrentPrice(AutoFallback):

    def get(self, crypto, fiat):
        if crypto.lower() == fiat.lower():
            return (1.0, 'math')

        return self._try_each_service('get_current_price', crypto=crypto, fiat=fiat)

    def no_service_msg(self, crypto, fiat):
        return "Can not find price for %s->%s" % (crypto, fiat)


class AddressBalance(AutoFallback):

    def get(self, crypto, address=None, addresses=None, confirmations=1):
        kwargs = dict(crypto=crypto, confirmations=confirmations)

        if address:
            method_name = "get_balance"
            kwargs['address'] = address

        if addresses:
            method_name = "get_balance_multi"
            kwargs['addresses'] = addresses

        return self._try_each_service(method_name, **kwargs)

    def no_service_msg(self, crypto, address, confirmations=1):
        return "Could not get confirmed address balance for: %s" % crypto


class PushTx(AutoFallback):

    def get(self, crypto, tx_hex):
        return self._try_each_service("push_tx", crypto=crypto, tx_hex=tx_hex)

    def no_service_msg(self, crypto, tx_hex):
        return "Could not push this %s transaction." % crypto


class HistoricalPrice(object):
    """
    This one doesn't inherit from AutoFallback because there is only one
    historical price API service at the moment.
    """
    def __init__(self, responses=None, verbose=False):
        self.service = Quandl(responses, verbose=verbose)

    def get(self, crypto, fiat, at_time):
        crypto = crypto.lower()
        fiat = fiat.lower()

        if crypto != 'btc' and fiat != 'btc':
            # two external requests and some math is going to be needed.
            from_btc, source1, date1 = self.service.get_historical(crypto, 'btc', at_time)
            to_altcoin, source2, date2 = self.service.get_historical('btc', fiat, at_time)
            return (from_btc * to_altcoin), "%s x %s" % (source1, source2), date1
        else:
            return self.service.get_historical(crypto, fiat, at_time)

    @property
    def responses(self):
        return self.service.responses



from moneywagon import (
    get_unspent_outputs, get_current_price, get_optimal_fee, push_tx
)
from bitcoin import mktx, sign

def from_unit_to_satoshi(value, unit, verbose):
    """
    Convert a value to satoshis. units can be any fiat currency
    """
    if not unit or unit == 'satoshi':
        return value
    if unit == 'bitcoin' or unit == 'btc':
        return value * 1e8

    # assume fiat currency that we can convert
    convert = get_current_price('btc', unit, verbose=verbose)[0]
    return int(value / convert * 1e8)


class Transaction(object):
    def __init__(self, crypto, hex=None, paranoid=1, utxo_services=None, pushtx_services=None, verbose=False):
        if not crypto.lower() == 'btc':
            raise ValueError("Transaction only supports BTC at this time")

        self.paranoid = paranoid
        self.crypto = crypto
        self.fee_satoshi = 10000
        self.outs = []
        self.ins = []
        self.verbose = verbose

        self.utxo_services = utxo_services or []
        self.pushtx_services = pushtx_services or []

        if hex:
            self.hex = hex

    def add_raw_inputs(self, inputs, private_key=None):
        """
        Add a set of utxo's to this transaction. This method is better to use if you
        want more fine control of which inputs get added to a transaction.
        `inputs` is a list of "unspent outputs" (they were 'outputs' to previous transactions,
          and 'inputs' to subsiquent transactions).

        `private_key` - All inputs will be signed by the passed in private key.
        """
        for i in inputs:
            self.ins.append(dict(input=i, private_key=private_key))
            self.change_address = i['address']

    def _get_utxos(self, address):
        """
        Using the service fallback engine, get utxos from remote service.
        """
        return get_unspent_outputs(
            self.crypto, address, services=self.utxo_services,
            paranoid=self.paranoid, verbose=self.verbose
        )

    def add_inputs_from_address(self, address, private_key=None, amount='all'):
        """
        Make call to external service to get inputs from an address.
        `amount` is the amount of [currency] worth of inputs to add from this address.
          pass in 'all' (the default) to use *all* inputs found for this address.
        """
        self.private_key = private_key
        self.change_address = address

        total_added = 0
        ins = []
        for utxo in self._get_utxos(address):
            if (amount == 'all' or total_added < amount):
                self.ins.append(
                    dict(input=utxo, private_key=private_key)
                )
                total_added += utxo['amount']

    def total_input_satoshis(self):
        """
        Add up all the satoshis coming from all input tx's.
        """
        just_inputs = [x['input'] for x in self.ins]
        return sum([x['amount'] for x in just_inputs])

    def add_output(self, address, value, unit=None):
        """
        Add an output (a person who will receive funds via this tx)
        """
        value_satoshi = from_unit_to_satoshi(value, unit, self.verbose)
        self.outs.append({
            'address': address,
            'value': value_satoshi
        })

    def fee(self, value, unit=None):
        """
        Set the miner fee, if unit is not set, assumes value is satoshi
        """
        if value == 'optimal':
            self.fee_satoshi = 'optimal'
        else:
            self.fee_satoshi = from_unit_to_satoshi(value, unit, self.verbose)

    def estimate_size(self):
        """
        Estimate how many bytes this transaction will be by countng inputs
        and outputs.
        Formula taken from: http://bitcoin.stackexchange.com/a/3011/18150
        """
        return len(self.outs) * 148 + 34 * len(self.ins) + 10

    def get_hex(self, signed=True):
        """
        Given all the data the user has given so far, make the hex using pybitcointools
        """
        total_ins = self.total_input_satoshis()
        total_outs = sum([x['value'] for x in self.outs])

        fee = self.fee_satoshi
        if fee == 'optimal':
            # makes call to external service to get optimal fee
            fee = get_optimal_fee(self.crypto, self.estimate_size(), 0, verbose=self.verbose)

        change_satoshi = total_ins - (total_outs + fee)

        if change_satoshi < 0:
            raise ValueError("Input amount must be more than all Output amounts. You need more bitcoin.")

        ins = [x['input'] for x in self.ins]

        tx = mktx(ins, self.outs + [{'address': self.change_address, 'value': change_satoshi}])

        if signed:
            for i, input_data in enumerate(self.ins):
                if not input_data['private_key']:
                    raise Exception("Can't sign transaction, missing private key")
                tx = sign(tx, i, input_data['private_key'])

        return tx

    def push(self):
        return push_tx(
            self.crypto, self.get_hex(), services=self.pushtx_services,
            verbose=self.verbose
        )



from .core import Service, NoService, NoData, SkipThisService, currency_to_protocol
import arrow

class Bitstamp(Service):
    supported_cryptos = ['btc']

    def get_current_price(self, crypto, fiat):
        if fiat.lower() != 'usd':
            raise SkipThisService('Bitstamp only does USD->BTC')

        url = "https://www.bitstamp.net/api/ticker/"
        response = self.get_url(url).json()
        return (float(response['last']), 'bitstamp')


class BlockCypher(Service):
    supported_cryptos = ['btc', 'ltc', 'uro']

    def get_balance(self, crypto, address, confirmations=1):
        url = "http://api.blockcypher.com/v1/%s/main/addrs/%s" % (crypto, address)
        response = self.get_url(url)
        return response.json()['balance'] / 1.0e8


class BlockSeer(Service):
    """
    This service has no publically documented API, this code was written
    from looking through chrome dev toolbar.
    """
    supported_cryptos = ['btc']

    def get_balance(self, crypto, address, confirmations=1):
        url = "https://www.blockseer.com/api/addresses/%s" % address
        return self.get_url(url).json()['data']['balance'] / 1e8

    def get_transactions(self, crypo, address):
        url = "https://www.blockseer.com/api/addresses/%s/transactions?filter=all" % address
        transactions = []
        for tx in self.get_url(url).json()['data']['address']['transactions']:
            transactions.append(dict(
                date=arrow.get(tx['time']).datetime,
                amount=tx['delta'] / 1e8,
                txid=tx['hash'],
            ))
        return transactions


class SmartBitAU(Service):
    base_url = "https://api.smartbit.com.au/v1/blockchain"
    supported_cryptos = ['btc']

    def get_balance(self, crypto, address, confirmations=1):
        url = "%s/address/%s" % (self.base_url, address)
        r = self.get_url(url).json()

        confirmed = float(r['address']['confirmed']['balance'])
        if confirmations > 1:
            return confirmed
        else:
            return confirmed + float(r['address']['unconfirmed']['balance'])

    def get_balance_muti(self, crypto, addresses, confirmations=1):
        url = "%s/address/%s" % (self.base_url, ",".join(addresses))
        response = self.get_url(url).json()

        ret = {}
        for data in response['addresses']:
            bal = float(data['confirmed']['balance'])
            if confirmations == 0:
                bal += float(data['unconfirmed']['balance'])
            ret[data['address']] = bal

        return ret

    def get_transactions(self, crypto, address, confirmations=1):
        url = "%s/address/%s" % (self.base_url, address)
        transactions = []
        for tx in self.get_url(url).json()['address']['transactions']:
            out_amount = sum(float(x['value']) for x in tx['outputs'] if address in x['addresses'])
            in_amount = sum(float(x['value']) for x in tx['inputs'] if address in x['addresses'])
            transactions.append(dict(
                amount=out_amount - in_amount,
                date=arrow.get(tx['time']).datetime,
                fee=float(tx['fee']),
                txid=tx['txid'],
                confirmations=tx['confirmations'],
            ))
        return transactions

    def get_unspent_outputs(self, crypto, address, confirmations=1):
        url = "%s/address/%s/unspent" % (self.base_url, address)
        utxos = []
        for utxo in self.get_url(url).json()['unspent']:
            utxos.append(dict(
                amount=utxo['value_int'],
                output="%s:%s" % (utxo['txid'], utxo['n']),
                address=address,
            ))
        return utxos

    def push_tx(self, crypto, tx_hex):
        """
        This method is untested.
        """
        url = "%s/pushtx" % self.base_url
        return self.post_url(url, {'hex': tx_hex}).content

    def get_mempool(self):
        url = "%s/transactions/unconfirmed?limit=1000" % self.base_url
        txs = []
        for tx in self.get_url(url).json()['transactions']:
            txs.append(dict(
                first_seen=arrow.get(tx['first_seen']).datetime,
                size=tx['size'],
                txid=tx['txid'],
                fee=float(tx['fee']),
            ))
        return txs


class Blockr(Service):
    supported_cryptos = ['btc', 'ltc', 'ppc', 'mec', 'qrk', 'dgc', 'tbtc']

    def get_balance(self, crypto, address, confirmations=1):
        url = "http://%s.blockr.io/api/v1/address/info/%s" % (crypto, address)
        response = self.get_url(url)
        return response.json()['data']['balance']

    def get_transactions(self, crypto, address, confirmations=1):
        url = 'http://%s.blockr.io/api/v1/address/txs/%s' % (crypto, address)
        response = self.get_url(url)

        transactions = []
        for tx in response.json()['data']['txs']:
            transactions.append(dict(
                date=arrow.get(tx['time_utc']).datetime,
                amount=tx['amount'],
                txid=tx['tx'],
                confirmations=tx['confirmations'],
            ))
        return transactions

    def get_unspent_outputs(self, crypto, address, confirmations=1):
        url = "http://%s.blockr.io/api/v1/address/unspent/%s" % (crypto, address)
        utxos = []
        for utxo in self.get_url(url).json()['data']['unspent']:
            utxos.append(dict(
                amount=currency_to_protocol(utxo['amount']),
                address=address,
                output="%s:%s" % (utxo['tx'], utxo['n']),
                confirmations=utxo['confirmations']
            ))
        return utxos

    def push_tx(self, crypto, tx_hex):
        url = "http://%s.blockr.io/api/v1/tx/push" % crypto
        response = self.post_url(url, {'tx': tx_hex})
        return response.json()['data']

    def get_block(self, crypto, block_hash='', block_number='', latest=False):
        url ="http://%s.blockr.io/api/v1/block/info/%s%s%s" % (
            crypto, block_hash, block_number, 'latest' if latest else ''
        )
        r = self.get_url(url).json()['data']
        return dict(
            block_number=r['nb'],
            confirmations=r['confirmations'],
            time=arrow.get(r['time_utc']).datetime,
            sent_value=r['vout_sum'],
            total_fees=float(r['fee']),
            mining_difficulty=r['difficulty'],
            size=r['size'],
            hash=r['hash'],
            merkle_root=r['merkleroot'],
            previous_hash=r['prev_block_hash'],
            next_hash=r['next_block_hash'],
            tx_count=r['nb_txs'],
        )


class Toshi(Service):
    url = "https://bitcoin.toshi.io/api/v0"

    supported_cryptos = ['btc']

    def get_balance(self, crypto, address, confirmations=1):
        url = "%s/addresses/%s" % (self.url, address)
        response = self.get_url(url).json()
        return response['balance'] / 1e8

    def get_transactions(self, crypto, address, confirmations=1):
        """
        This call also returns unconfirmed transactions.
        """
        url = "%s/addresses/%s/transactions" % (self.url, address)
        response = self.get_url(url).json()

        if confirmations == 0:
            to_iterate = response['transactions'] + response['unconfirmed_transactions']
        else:
            to_iterate = response['transactions']

        transactions = []
        for tx in to_iterate:
            transactions.append(dict(
                amount=sum([x['amount'] / 1e8 for x in tx['outputs'] if address in x['addresses']]),
                txid=tx['hash'],
                date=arrow.get(tx['block_time']).datetime,
                confirmations=tx['confirmations']
            ))
        return transactions

    def push_tx(self, crypto, tx_hex):
        url = "%s/transactions/%s" % (self.url, tx_hex)
        return self.get_url(url).json()['hash']

    def get_block(self, crypto, block_hash='', block_number='', latest=False):
        if latest:
            url = "%s/blocks/latest" % self.url
        else:
            url = "%s/blocks/%s%s" % (self.url, block_hash, block_number)

        r = self.get_url(url).json()
        return dict(
            block_number=r['height'],
            confirmations=r['confirmations'],
            time=arrow.get(r['time']).datetime,
            sent_value=r['total_out'] / 1e8,
            total_fees=r['fees'] / 1e8,
            mining_difficulty=r['difficulty'],
            size=r['size'],
            hash=r['hash'],
            merkle_root=r['merkle_root'],
            previous_hash=r['previous_block_hash'],
            next_hash=r['next_blocks'][0]['hash'] if len(r['next_blocks']) else None,
            txids=sorted(r['transaction_hashes']),
            tx_count=len(r['transaction_hashes'])
        )


class BTCE(Service):
    def get_current_price(self, crypto, fiat):
        pair = "%s_%s" % (crypto.lower(), fiat.lower())
        url = "https://btc-e.com/api/3/ticker/" + pair
        response = self.get_url(url).json()
        return (response[pair]['last'], 'btc-e')


class Cryptonator(Service):
    def get_current_price(self, crypto, fiat):
        pair = "%s-%s" % (crypto, fiat)
        url = "https://www.cryptonator.com/api/ticker/%s" % pair
        response = self.get_url(url).json()
        return float(response['ticker']['price']), 'cryptonator'


class Winkdex(Service):
    supported_cryptos = ['btc']

    def get_current_price(self, crypto, fiat):
        if fiat != 'usd':
            raise SkipThisService("winkdex is btc->usd only")
        url = "https://winkdex.com/api/v0/price"
        return self.get_url(url).json()['price'] / 100.0, 'winkdex'


class BlockStrap(Service):
    """
    Documentation here: http://docs.blockstrap.com/en/api/
    """
    domain = 'http://api.blockstrap.com'
    supported_cryptos = ['btc', 'ltc', 'drk', 'doge']

    def get_balance(self, crypto, address, confirmations=None):
        url = "%s/v0/%s/address/id/%s" % (self.domain, crypto, address)
        response = self.get_url(url).json()
        return response['data']['address']['balance'] / 1e8

    def push_tx(self, crypto, tx_hex):
        url = "%s/v0/%s/transaction/relay/%s" % (self.domain, crypto, tx_hex)
        return self.get_url(url)['data']['id']

    def get_transactions(self, crypto, address):
        url = "%s/v0/%s/address/transactions/%s" % (self.domain, crypto, address)
        txs = []
        for tx in self.get_url(url).json()['data']['address']['transactions']:
            s_amount = tx['tx_address_input_value'] or tx['tx_address_output_value'] * -1
            txs.append(dict(
                date=arrow.get(tx['block_time']).datetime,
                amount=s_amount / 1e8,
                confirmations=tx['confirmations'],
                txid=tx['id'],
            ))
        return txs

    def get_unspent_outputs(self, crypto, address):
        url = "%s/v0/%s/address/unspents/%s" % (self.domain, crypto, address)
        utxos = []
        for utxo in self.get_url(url).json()['data']['address']['transactions']:
            utxos.append(dict(
                amount=utxo['tx_address_value'],
                address=address,
                output="%s:%s" % (utxo['id'].lower(), utxo['tx_address_tx_pos']),
                confirmations=utxo['confirmations']
            ))
        return utxos

    def get_block(self, crypto, block_hash='', block_number='', latest=False):
        if block_hash:
            url = "%s/v0/%s/block/id/%s" % (self.domain, crypto, block_hash)
            b = "block"
        elif block_number:
            url = "%s/v0/%s/block/height/%s" % (self.domain, crypto, block_number)
            b = "blocks"
        elif latest:
            url = "%s/v0/%s/block/latest" % (self.domain, crypto)
            b = 'block'

        r = self.get_url(url).json()['data'][b]

        if block_number:
            r = r[0]

        return dict(
            block_number=r['height'],
            confirmations=r["confirmations"],
            time=arrow.get(r['time_display']).datetime,
            sent_value=r["output_value"] / 1e8,
            total_fees=r["fees"] / 1e8,
            size=r['size'],
            hash=r['id'].lower(),
            merkle_root=r['merkel_root'].lower(),
            previous_hash=r['prev_block_id'].lower(),
            next_hash=r['next_block_id'].lower(),
            txids=sorted(x['id'] for x in r['transactions']),
            tx_count=len(r['transactions'])
        )


class ChainSo(Service):
    base_url = "https://chain.so/api/v2"
    supported_cryptos = ['doge', 'btc', 'ltc']

    def get_current_price(self, crypto, fiat):
        url = "%s/get_price/%s/%s" % (self.base_url, crypto, fiat)
        resp = self.get_url(url).json()
        items = resp['data']['prices']
        if len(items) == 0:
            raise SkipThisService("Chain.so can't get price for %s/%s" % (crypto, fiat))
        return float(items[0]['price']), "%s via Chain.so" % items[0]['exchange']

    def get_balance(self, crypto, address, confirmations=1):
        url = "%s/get_address_balance/%s/%s/%s" % (
            self.base_url, crypto, address, confirmations
        )
        response = self.get_url(url)
        return float(response.json()['data']['confirmed_balance'])

    def get_transactions(self, crypto, address, confirmations=1):
        url = "%s/get_tx_received/%s/%s" % (self.base_url, crypto, address)
        response = self.get_url(url)

        transactions = []
        for tx in response.json()['data']['txs']:
            tx_cons = int(tx['confirmations'])
            if tx_cons < confirmations:
                continue
            transactions.append(dict(
                date=arrow.get(tx['time']).datetime,
                amount=float(tx['value']),
                txid=tx['txid'],
                confirmations=tx_cons,
            ))

        return transactions

    def get_unspent_outputs(self, crypto, address, confirmations=1):
        url = "%s/get_tx_unspent/%s/%s" %(self.base_url, crypto, address)
        utxos = []
        for utxo in self.get_url(url).json()['data']['txs']:
            utxos.append(dict(
                amount=currency_to_protocol(utxo['value']),
                address=address,
                output="%s:%s" % (utxo['txid'], utxo['output_no']),
                confirmations=utxo['confirmations']
            ))
        return utxos


    def push_tx(self, crypto, tx_hex):
        url = "%s/send_tx/%s" % (self.base_url, crypto)
        resp = self.post_url(url, {'tx_hex': tx_hex})
        return resp.json()['txid']

    def get_block(self, crypto, block_number='', block_hash='', latest=False):
        if latest:
            raise SkipThisService("This service can't get block by latest")
        else:
            url = "%s/block/%s/%s%s" % (
                self.base_url, crypto, block_number, block_hash
            )
        r = self.get_url(url).json()['data']
        return dict(
            block_number=r['block_no'],
            confirmations=r['confirmations'],
            time=arrow.get(r['time']).datetime,
            sent_value=float(r['sent_value']),
            total_fees=float(r['fee']),
            mining_difficulty=float(r['mining_difficulty']),
            size=r['size'],
            hash=r['blockhash'],
            merkle_root=r['merkleroot'],
            previous_hash=r['previous_blockhash'],
            next_hash=r['next_blockhash'],
            txids=sorted([t['txid'] for t in r['txs']])
        )


class CoinPrism(Service):
    base_url = "https://api.coinprism.com/v1"
    supported_cryptos = ['btc']

    def get_balance(self, crypto, address, confirmations=None):
        url = "%s/addresses/%s" % (self.base_url, address)
        resp = self.get_url(url).json()
        return resp['balance'] / 1e8

    def get_transactions(self, crypto, address):
        url = "%s/addresses/%s/transactions" % (self.base_url, address)
        transactions = []
        for tx in self.get_url(url).json():
            transactions.append(dict(
                amount=sum([x['value'] / 1e8 for x in tx['outputs'] if address in x['addresses']]),
                txid=tx['hash'],
                date=arrow.get(tx['block_time']).datetime,
                confirmations=tx['confirmations']
            ))

        return transactions

    def get_unspent_outputs(self, crypto, address):
        url = "%s/addresses/%s/unspents" % (self.base_url, address)
        transactions = []
        for tx in self.get_url(url).json():
            if address in tx['addresses']:
                transactions.append(dict(
                    amount=tx['value'],
                    address=address,
                    output="%s:%s" % (tx['transaction_hash'], tx['output_index']),
                    confirmations=tx['confirmations']
                ))

        return transactions

    def push_tx(self, crypto, tx_hex):
        """
        Note: This one has not been tested yet.
        http://docs.coinprism.apiary.io/#reference/transaction-signing-and-broadcasting/push-a-signed-raw-transaction-to-the-network/post
        """
        url = "%s/sendrawtransaction"
        return self.post_url(url, tx_hex).content


class BitEasy(Service):
    """
    Most functions from this servie require an API key. therefore only
    address balance is supported at this time.
    """
    supported_cryptos = ['btc']

    def get_balance(self, crypto, address, confirmations=1):
        url = "https://api.biteasy.com/blockchain/v1/addresses/" + address
        response = self.get_url(url)
        return response.json()['data']['balance'] / 1e8


class BlockChainInfo(Service):
    supported_cryptos = ['btc']

    def get_balance(self, crypto, address, confirmations=1):
        url = "https://blockchain.info/address/%s?format=json" % address
        response = self.get_url(url)
        return float(response.json()['final_balance']) * 1e-8

    def get_unspent_outputs(self, crypto, address, confirmations=1):
        url = "https://blockchain.info/unspent?address=%s" % address
        utxos = []
        for utxo in self.get_url(url).json()['unspent_outputs']:
            if utxo['confirmations'] < confirmations:
                continue # don't return if too few confirmations

            utxos.append(dict(
                output="%s:%s" % (utxo['tx_hash_big_endian'], utxo['tx_output_n']),
                amount=utxo['value'],
                address=address,
            ))
        return utxos


##################################

class BitcoinAbe(Service):
    supported_cryptos = ['btc']
    base_url = "http://bitcoin-abe.info/chain/Bitcoin"

    def get_balance(self, crypto, address, confirmations=1):
        url = self.base_url + "/q/addressbalance/" + address
        response = self.get_url(url)
        return float(response.content)


class LitecoinAbe(BitcoinAbe):
    supported_cryptos = ['ltc']
    base_url = "http://bitcoin-abe.info/chain/Litecoin"


class NamecoinAbe(BitcoinAbe):
    supported_cryptos = ['nmc']
    base_url = "http://bitcoin-abe.info/chain/Namecoin"


class DogeChainInfo(BitcoinAbe):
    supported_cryptos = ['doge']
    base_url = "https://dogechain.info/chain/Dogecoin"


class AuroraCoinEU(BitcoinAbe):
    supported_cryptos = ['aur']
    base_url = 'http://blockexplorer.auroracoin.eu/chain/AuroraCoin'


class Atorox(BitcoinAbe):
    supported_cryptos = ['aur']
    base_url = "http://auroraexplorer.atorox.net/chain/AuroraCoin"

##################################

class FeathercoinCom(Service):
    supported_cryptos = ['ftc']

    def get_balance(self, crypto, address, confirmations=1):
        url= "http://api.feathercoin.com/?output=balance&address=%s&json=1" % address
        response = self.get_url(url)
        return float(response.json()['balance'])


class NXTPortal(Service):
    supported_cryptos = ['nxt']

    def get_balance(self, crypto, address, confirmations=1):
        url='http://nxtportal.org/nxt?requestType=getAccount&account=' + address
        response = self.get_url(url)
        return float(response.json()['balanceNQT']) * 1e-8

    def get_transactions(self, crypto, address):
        url = 'http://nxtportal.org/transactions/account/%s?num=50' % address
        response = self.get_url(url)
        transactions = []
        for tx in txs:
            transactions.append(dict(
                date=arrow.get(tx['time']).datetime,
                amount=tx['value'],
                txid=tx['txid'],
                confirmations=tx['confirmations'],
            ))

        return transactions


class CryptoID(Service):
    supported_cryptos = [
        'dash', 'bc', 'bay', 'block', 'cann', 'uno', 'vrc', 'xc', 'uro', 'aur',
        'pot', 'cure', 'arch', 'swift', 'karm', 'dgc', 'lxc', 'sync', 'byc',
        'pc', 'fibre', 'i0c', 'nobl', 'gsx', 'flt', 'ccn', 'rlc', 'rby', 'apex',
        'vior', 'ltcd', 'zeit', 'carbon', 'super', 'dis', 'ac', 'vdo', 'ioc',
        'xmg', 'cinni', 'crypt', 'excl', 'mne', 'seed', 'qslv', 'maryj', 'key',
        'oc', 'ktk', 'voot', 'glc', 'drkc', 'mue', 'gb', 'piggy', 'jbs', 'grs',
        'icg', 'rpc', ''
    ]

    def get_balance(self, crypto, address, confirmations=1):
        url ="http://chainz.cryptoid.info/%s/api.dws?q=getbalance&a=%s" % (crypto, address)
        return float(self.get_url(url).content)


class CryptapUS(Service):
    supported_cryptos = [
        'nmc', 'wds', 'ber', 'scn', 'sc0', 'wdc', 'nvc', 'cas', 'myr'
    ]
    def get_balance(self, crypto, address, confirmations=1):
        url = "http://cryptap.us/%s/explorer/q/addressbalance/%s" % (crypto, address)
        return float(self.get_url(url).content)


class BTER(Service):
    def get_current_price(self, crypto, fiat):
        url_template = "http://data.bter.com/api/1/ticker/%s_%s"
        url = url_template % (crypto, fiat)

        response = self.get_url(url).json()

        if response['result'] == 'false': # bter api returns this as string
            # bter doesn't support this pair, we need to make 2 calls and
            # do the math ourselves. The extra http request isn't a problem because
            # of caching. BTER only has USD, BTC and CNY
            # markets, so any other fiat will likely fail.

            url = url_template % (crypto, 'btc')
            response = self.get_url(url)
            altcoin_btc = float(response['last'])

            url = url_template % ('btc', fiat)
            response = self.get_url(url)
            btc_fiat = float(response['last'])

            return (btc_fiat * altcoin_btc), 'bter (calculated)'

        return float(response['last'] or 0), 'bter'


class CoinSwap(Service):
    def get_current_price(self, crypto, fiat):
        chunk = ("%s/%s" % (crypto, fiat)).upper()
        url = "https://api.coin-swap.net/market/stats/%s" % chunk
        response = self.get_url(url).json()
        return float(response['lastprice']), 'coin-swap'


class ExCoIn(Service):
    # decommissioned
    def get_current_price(self, crypto, fiat):
        url = "https://api.exco.in/v1/exchange/%s/%s" % (fiat, crypto)
        response = self.get_url(url).json()
        return float(response['last_price']), 'exco.in'

################################################

class BitpayInsight(Service):
    supported_cryptos = ['btc']
    domain = "http://insight.bitpay.com"

    def get_balance(self, crypto, address, confirmations=1):
        url = "%s/api/addr/%s/balance" % (self.domain, address)
        return float(self.get_url(url).content) / 1e8

    def get_transactions(self, crypto, address):
        url = "%s/api/txs/?address=%s" % (self.domain, address)
        response = self.get_url(url)

        transactions = []
        for tx in response.json()['txs']:
            my_outs = [
                float(x['value']) for x in tx['vout'] if address in x['scriptPubKey']['addresses']
            ]
            transactions.append(dict(
                amount=sum(my_outs),
                date=arrow.get(tx['time']).datetime,
                txid=tx['txid'],
                confirmations=tx['confirmations'],
            ))

        return transactions

    def get_unspent_outputs(self, crypto, address, confirmations=1):
        url = "%s/api/addr/%s/utxo?noCache=1" % (self.domain, address)
        utxos = []
        for utxo in self.get_url(url).json():
            utxos.append(dict(
                output="%s:%s" % (utxo['txid'], utxo['vout']),
                amount=currency_to_protocol(utxo['amount']),
                confirmations=utxo['confirmations'],
                address=address
            ))
        return utxos

    def get_block(self, crypto, block_number='', block_hash='', latest=False):
        if block_number:
            url = "%s/api/block-index/%s" % (self.domain, block_number)
        elif block_hash:
            url = "%s/api/block/%s" % (self.domain, block_hash)
        elif latest:
            raise ValueError("Can't get current block.")

        r = self.get_url(url).json()
        return dict(
            block_number=r['block_no'],
            confirmations=r['confirmations'],
            time=arrow.get(r['time']).datetime,
            sent_value=float(r['sent_value']),
            total_fees=float(r['fee']),
            mining_difficulty=float(r['mining_difficulty']),
            size=r['size'],
            hash=r['blockhash'],
            merkle_root=r['merkleroot'],
            previous_hash=r['previous_blockhash'],
            next_hash=r['next_blockhash'],
            txids=[t['txid'] for t in r['txs']],
            tx_count=len(r['txs'])
        )


class TheBitInfo(BitpayInsight):
    supported_cryptos = ['btc']
    domain = "http://insight.thebit.info/"


class MYRCryptap(BitpayInsight):
    supported_cryptos = ['myr']
    domain = "http://insight-myr.cryptap.us/"


class BirdOnWheels(BitpayInsight):
    supported_cryptos = ['myr']
    domain = "http://birdonwheels5.no-ip.org:3000"


class ThisIsVTC(BitpayInsight):
    supported_cryptos = ['vtc']
    domain = "http://explorer.thisisvtc.com"


class ReddcoinCom(BitpayInsight):
    supported_cryptos = ['rdd']
    domain = "http://live.reddcoin.com"


class FTCe(BitpayInsight):
    supported_cryptos = ['ftc']
    domain = "http://block.ftc-c.com"


class CoinTape(Service):
    supported_cryptos = ['btc']

    def get_optimal_fee(self, crypto, tx_bytes, acceptable_block_delay=0):
        url = "http://www.cointape.com/fees"
        response = self.get_url(url).json()
        for sample in response['fees']:
            if sample['maxDelay'] <= acceptable_block_delay:
                rate = sample['maxFee']
                return tx_bytes * rate


ALL_SERVICES = [
    Bitstamp, BlockCypher, Blockr, BTCE, Cryptonator, Winkdex,
    BitEasy, BlockChainInfo, BitcoinAbe, LitecoinAbe, NamecoinAbe, DogeChainInfo,
    AuroraCoinEU, Atorox, FeathercoinCom, NXTPortal, CryptoID,
    CryptapUS, BTER, CoinSwap, ChainSo, BlockStrap, CoinPrism, BlockSeer,

    BitpayInsight, ThisIsVTC, BirdOnWheels, MYRCryptap, ReddcoinCom, FTCe,

    CoinTape
]



from datetime import datetime
from .services import *

crypto_data = {
    'btc': {
        'name': 'Bitcoin',
        'genesis_date': datetime(2009, 1, 12),
        'services': {
            'current_price': [
                Bitstamp, Winkdex, BTCE, BTER, Cryptonator, ChainSo
            ],
            'address_balance': [
                Toshi, BlockCypher, Blockr, BlockStrap, ChainSo, BlockSeer,
                BitEasy, SmartBitAU, BlockChainInfo, BitcoinAbe,
                BitpayInsight, CoinPrism
            ],
            'historical_transactions': [
                Blockr, Toshi, SmartBitAU, ChainSo, CoinPrism, BlockSeer,
                BitpayInsight
            ],
            'push_tx': [
                BlockChainInfo, Blockr, BlockStrap, ChainSo, CoinPrism,
                TheBitInfo, BitpayInsight
            ],
            'unspent_outputs': [
                Blockr, BitpayInsight, SmartBitAU, BlockChainInfo, CoinPrism, ChainSo,
                BlockStrap, TheBitInfo
            ],
            'get_block': [
                BitpayInsight,Toshi, Blockr, ChainSo, BlockStrap, TheBitInfo
            ]
        },
    },
    'ltc': {
        'name': 'Litecoin',
        'genesis_date': datetime(2011, 10, 7),
        'services': {
            'current_price': [
                BTCE, ChainSo, BTER, Cryptonator
            ],
            'address_balance': [
                BlockCypher, Blockr, BlockStrap, LitecoinAbe, ChainSo
            ],
            'historical_transactions': [
                Blockr, ChainSo
            ],
            'push_tx': [
                Blockr, BlockStrap, ChainSo
            ],
            'unspent_outputs': [
                BlockStrap, ChainSo, Blockr
            ],
            'get_block': [
                ChainSo, Blockr, BlockStrap
            ]
        },
    },
    'ppc': {
        'name': 'Peercoin',
        'genesis_date': datetime(2012, 8, 19),
        'services': {
            'current_price': [
                BTCE, ChainSo, BTER, Cryptonator
            ],
            'address_balance': [
                Blockr
            ],
            'historical_transactions': [
                Blockr
            ],
            'push_tx': [
                Blockr
            ],
            'unspent_outputs': [
                Blockr
            ],
            'get_block': [
                Blockr
            ]
        },
    },
    'doge': {
        'name': 'Dogecoin',
        'genesis_date': datetime(2013, 12, 6),
        'services': {
            'current_price': [
                BTER, ChainSo, Cryptonator
            ],
            'address_balance': [
                DogeChainInfo, BlockStrap
            ],
            'historical_transactions': [
                ChainSo
            ],
            'push_tx': [
                BlockStrap, ChainSo
            ],
            'unspent_outputs': [
                ChainSo, BlockStrap
            ],
            'get_block': [
                BlockStrap, ChainSo
            ]
        },
    },
    'nxt': {
        'name': 'Nxt',
        'genesis_date': datetime(2013, 10, 29),
        'services': {
            'current_price': [
                ChainSo, Cryptonator
            ],
            'address_balance': [
                NXTPortal
            ],
            'historical_transactions': [
                NXTPortal
            ],
            'push_tx': [

            ],
            'unspent_outputs': [

            ],
            'get_block': [

            ]
        },
    },
    'myr': {
        'name': 'MyriadCoin',
        'genesis_date': datetime(2014, 2, 23),
        'services': {
            'current_price': [
                Cryptonator, ChainSo
            ],
            'address_balance': [
                MYRCryptap, CryptapUS, BirdOnWheels
            ],
            'historical_transactions': [
                MYRCryptap, BirdOnWheels
            ],
            'push_tx': [
                MYRCryptap
            ],
            'unspent_outputs': [
                MYRCryptap
            ],
            'get_block': [
                MYRCryptap
            ]
        },
    },
    'vtc': {
        'name': 'Vertcoin',
        'genesis_date': datetime(2014, 1, 8),
        'services': {
            'current_price': [
                Cryptonator, ChainSo
            ],
            'address_balance': [
                ThisIsVTC
            ],
            'historical_transactions': [
                ThisIsVTC
            ],
            'push_tx': [
                ThisIsVTC
            ],
            'unspent_outputs': [
                ThisIsVTC
            ],
            'get_block': [
                ThisIsVTC
            ]
        },
    },
    'ftc': {
        'name': 'Feathercoin',
        'genesis_date': datetime(2013, 4, 16),
        'services': {
            'current_price': [
                ChainSo, Cryptonator
            ],
            'address_balance': [
                FTCe
            ],
            'historical_transactions': [
                FTCe
            ],
            'push_tx': [
                FTCe
            ],
            'unspent_outputs': [
                FTCe
            ],
            'get_block': [
                FTCe
            ]
        },
    },
    'dash': {
        'name': 'Dash',
        'genesis_date': datetime(2014, 1, 19),
        'services': {
            'current_price': [
                ChainSo, Cryptonator
            ],
            'address_balance': [
                CryptoID
            ],
            'historical_transactions': [

            ],
            'push_tx': [

            ],
            'unspent_outputs': [

            ],
            'get_block': [

            ]
        },
    },

    'rdd': {
        'name': 'Reddcoin',
        'genesis_date': datetime(2014, 1, 20),
        'services': {
            'current_price': [
                ChainSo, Cryptonator
            ],
            'address_balance': [
                ReddcoinCom
            ],
            'historical_transactions': [
                ReddcoinCom
            ],
            'push_tx': [
                ReddcoinCom
            ],
            'unspent_outputs': [
                ReddcoinCom
            ],
            'get_block': [
                ReddcoinCom
            ]
        },
    },

    # TEMPLATE
    '': {
        'name': '',
        'genesis_date': datetime(2014, 1, 2),
        'services': {
            'current_price': [

            ],
            'address_balance': [

            ],
            'historical_transactions': [

            ],
            'push_tx': [

            ],
            'unspent_outputs': [

            ],
            'get_block': [

            ]
        },
    },

    # these ones below need to be modified to match the above format
    # taken from http://coinwik.org/List_of_all_DCs
    # sorted by date released, copied over on Nov 27, 2014

    'nmc': ['Namecoin', datetime(2011, 4, 18)],
    'tk':  ['TimeKoin', datetime(2011, 5, 27)],
    'dvc': ['DevCoin', datetime(2011, 8, 5)],
    'ixc': ['IxCoin', datetime(2011, 8, 10)],
    'bcn': ['Bytecoin', datetime(2012, 7, 4)],
    'bte': ['Bytecoin', datetime(2012, 7, 4)],
    'bqc': ['BBQCoin', datetime(2012, 7, 15)],
    'str': ['StarCoin', datetime(2012, 7, 19)],

    'nvc': ['NovaCoin', datetime(2012, 10, 1)],
    'trc': ['TerraCoin', datetime(2012, 10, 26)],
    'frc': ['Freicoin', datetime(2012, 12, 21)],

    'btb': ['BitBar', datetime(2013, 4, 1)],
    'mnc': ['MinCoin', datetime(2013, 4, 3)],
    'net': ['NetCoin', datetime(2013, 4, 7)],
    'nan': ['NanoToken', datetime(2013, 4, 8)],

    'xrp': ['Ripple', datetime(2011, 3, 1)],
    'pxc': ['PhoenixCoin', datetime(2013, 5, 1)],
    'cnc': ['CHNCoin', datetime(2013, 5, 1)],
    'jkc': ['JunkCoin', datetime(2013, 5, 4)],
    'yac': ['YaCoin', datetime(2013, 5, 5)],
    'ryc': ['RoyalCoin', datetime(2013, 5, 9)],
    'onc': ['OneCoin', datetime(2013, 5, 9)],
    'frk': ['Franko', datetime(2013, 5, 11)],
    'gme': ['GameCoin', datetime(2013, 5, 12)],
    'pwc': ['PowerCoin', datetime(2013, 5, 13)],
    'elc': ['ElaCoin', datetime(2013, 5, 13)],
    'btg': ['BitGem', datetime(2013, 5, 16)],
    'dbl': ['Doubloons', datetime(2013, 5, 16)],
    'dgc': ['DigitalCoin', datetime(2013, 5, 18)],
    'nbl': ['Nibble', datetime(2013, 5, 18)],
    'pxc': ['PhenixCoin', datetime(2013, 5, 20)],
    'wdc': ['WorldCoin', datetime(2013, 5, 22)],
    'lky': ['LuckyCoin', datetime(2013, 5, 24)],
    'mem': ['MemeCoin', datetime(2013, 5, 27)],
    'hyc': ['HyperCoin', datetime(2013, 5, 28)],
    'sxc': ['SexCoin', datetime(2013, 5, 28)],
    'mex': ['MegaCoin', datetime(2013, 5, 29)],
    'fst': ['FastCoin', datetime(2013, 5, 29)],
    'amc': ['AmericanCoin', datetime(2013, 5, 29)],
    'ezc': ['EZCoin', datetime(2013, 5, 29)],
    'ifc': ['InfiniteCoin', datetime(2013, 6, 5)],
    'crc': ['CraftCoin', datetime(2013, 6, 5)],
    'anc': ['AnonCoin', datetime(2013, 6, 6)],
    'nrb': ['NoirBits', datetime(2013, 6, 6)],
    'clr': ['CopperLark', datetime(2013, 6, 6)],
    'rec': ['RealCoin', datetime(2013, 6, 6)],
    'sbc': ['StableCoin', datetime(2013, 6, 7)],
    'alf': ['AlphaCoin', datetime(2013, 6, 9)],
    'gld': ['GoldCoin', datetime(2013, 6, 11)],
    'arg': ['Argentum', datetime(2013, 6, 11)],
    'emd': ['Emerald', datetime(2013, 6, 16)],
    'xnc': ['XenCoin', datetime(2013, 6, 19)],
    'rch': ['Richcoin', datetime(2013, 6, 20)],
    'cap': ['BottleCaps', datetime(2013, 6, 23)],
    'nuc': ['Nucoin', datetime(2013, 6, 24)],
    'glc': ['GlobalCoin', datetime(2013, 6, 26)],
    'cgb': ['CryptogenicBullion', datetime(2013, 6, 27)],
    'cmc': ['CosmosCoin', datetime(2013, 6, 28)],
    'kgc': ['KrugerCoin', datetime(2013, 6, 30)],
    'red': ['RedCoin', datetime(2013, 6, 30)],
    'xpm': ['Primecoin', datetime(2013, 7, 7)],
    'glx': ['GalaxyCoin', datetime(2013, 7, 8)],
    'dmd': ['Diamond', datetime(2013, 7, 13)],
    'gdc': ['GrandCoin', datetime(2013, 7, 13)],
    'mst': ['MasterCoin (Hydro)', datetime(2013, 7, 13)],
    'gil': ['GilCoin', datetime(2013, 7, 14)],
    'elp': ['ElephantCoin', datetime(2013, 7, 15)],
    'flo': ['FlorinCoin', datetime(2013, 7, 17)],
    'csc': ['CasinoCoin', datetime(2013, 7, 18)],
    'qrk': ['Quark', datetime(2013, 7, 21)],
    'lbw': ['LEBOWSKIS', datetime(2013, 7, 21)],
    'hbn': ['Hobonickels', datetime(2013, 7, 24)],
    'nec': ['NeoCoin', datetime(2013, 7, 27)],
    'phs': ['PhilosopherStone', datetime(2013, 7, 28)],
    'msc': ['MasterCoin', datetime(2013, 7, 31)],
    'orb': ['OrbitCoin', datetime(2013, 7, 31)],
    'zet': ['ZetaCoin', datetime(2013, 8, 3)],
    'cent':['Pennies', datetime(2013, 8, 4)],
    'pyc': ['PayCoin', datetime(2013, 8, 6)],
    'zcc': ['ZCCoin', datetime(2013, 8, 7)],
    'cpr': ['CopperBars', datetime(2013, 8, 11)],
    'exc': ['Extremecoin', datetime(2013, 8, 17)],
    'adt': ['AndroidsTokens', datetime(2013, 8, 18)],
    'col': ['ColossusCoin', datetime(2013, 8, 22)],
    'i0c': ['I0Coin', datetime(2013, 8, 27)],
    'src': ['SecureCoin', datetime(2013, 8, 28)],
    'tgc': ['TigerCoin', datetime(2013, 9, 6)],
    'tek': ['TekCoin', datetime(2013, 9, 7)],
    'asc': ['AsicCoin', datetime(2013, 9, 11)],
    'lk7': ['Lucky7Coin', datetime(2013, 9, 15)],
    'tix': ['Tickets', datetime(2013, 9, 19)],
    'xjo': ['JouleCoin', datetime(2013, 9, 22)],
    'buk': ['CryptoBuck', datetime(2013, 9, 25)],
    'dem': ['eMark', datetime(2013, 10, 12)],
    'grc': ['GridCoin', datetime(2013, 10, 17)],
    'uno': ['Unobtanium', datetime(2013, 10, 17)],
    'spt': ['Spots', datetime(2013, 10, 18)],
    'tag': ['TagCoin', datetime(2013, 10, 24)],
    'bet': ['BetaCoin', datetime(2013, 10, 24)],
    'ffc': ['FireFlyCoin', datetime(2013, 10, 25)],
    'osc': ['OpenSourceCoin', datetime(2013, 11, 4)],
    'pts': ['Bitshares-PTS', datetime(2013, 11, 5)],
    'naut':['Nautiluscoin', datetime(2013, 11, 5)],
    'pts': ['ProtoShares', datetime(2013, 11, 5)],
    'apc': ['Applecoin', datetime(2013, 11, 19)],
    'cin': ['Cinnamoncoin', datetime(2013, 11, 21)],
    'ybc': ['YBCoin', datetime(2013, 11, 23)],

    'gra': ['Grain', datetime(2013, 12, 7)],
    'fz':  ['Frozen', datetime(2013, 12, 10)],
    'lot': ['LottoCoin', datetime(2013, 12, 12)],
    'mmc': ['Memorycoin', datetime(2013, 12, 15)],
    'dtc': ['Datacoin', datetime(2013, 12, 17)],
    'xiv': ['Xivra', datetime(2013, 12, 18)],
    'tips':['FedoraCoin', datetime(2013, 12, 21)],
    'ato': ['AtomCoin', datetime(2013, 12, 22)],
    'cat': ['CatCoin', datetime(2013, 12, 23)],
    'meow':['KittehCoin', datetime(2013, 12, 24)],
    'moon':['MoonCoin', datetime(2013, 12, 28)],
    'qqc': ['QQcoin', datetime(2013, 12, 28)],
    'rpc': ['RonPaulCoin', datetime(2013, 12, 29)],
    'vel': ['Velocitycoin', datetime(2013, 12, 29)],
    'eac': ['EarthCoin', datetime(2013, 12, 31)],
    'prt': ['Particle', datetime(2013, 12, 31)],
    'mona':['MonaCoin', datetime(2013, 12, 31)],
    'slr': ['SolarCoin', datetime(2014, 1, 1)],
    'xcp': ['Counterparty', datetime(2014, 1, 2)],
    'volt':['Electric', datetime(2014, 1, 2)],
    '42':  ['42Coin', datetime(2014, 1, 4)],
    'pnd': ['Pandacoin', datetime(2014, 1, 4)],
    'nut': ['NutCoin', datetime(2014, 1, 4)],
    'cach':['Cachecoin', datetime(2014, 1, 5)],
    'nyan':['Nyancoin', datetime(2014, 1, 5)],
    'smc': ['Smartcoin', datetime(2014, 1, 6)],
    'stc': ['StockCoin', datetime(2014, 1, 6)],
    'nobl':['Noblecoin', datetime(2014, 1, 7)],
    'ptc': ['PesetaCoin', datetime(2014, 1, 7)],
    'aph': ['AphroditeCoin', datetime(2014, 1, 7)],

    'rpd': ['Rapidcoin', datetime(2014, 1, 8)],
    'etok':['Etoken', datetime(2014, 1, 8)],
    'kdc': ['Klondike', datetime(2014, 1, 9)],
    'pir': ['PirateCoin', datetime(2014, 1, 9)],
    'dgb': ['Digibyte', datetime(2014, 1, 10)],
    'xc':  ['XCurrency', datetime(2014, 1, 10)],
    'bcx': ['BattleCoin', datetime(2014, 1, 10)],
    'icn': ['iCoin', datetime(2014, 1, 10)],
    'krn': ['Ekrona', datetime(2014, 1, 10)],
    'usde':['Unitary Status Dollar eCoin', datetime(2014, 1, 12)],
    'q2c': ['Qubitcoin', datetime(2014, 1, 12)],
    'zed': ['ZedCoin', datetime(2014, 1, 14)],
    'mtc': ['MarineCoin', datetime(2014, 1, 15)],
    'fox': ['FoxCoin', datetime(2014, 1, 17)],

    'tes': ['Teslacoin', datetime(2014, 1, 18)],
    'con': ['Coino', datetime(2014, 1, 18)],
    'plc': ['PolCoin', datetime(2014, 1, 19)],

    'cash':['CashCoin', datetime(2014, 1, 20)],
    'mry': ['MurrayCoin', datetime(2014, 1, 20)],
    'pot': ['Potcoin', datetime(2014, 1, 21)],
    'asr': ['AstroCoin', datetime(2014, 1, 21)],
    'pmc': ['Premine', datetime(2014, 1, 23)],
    'leaf':['LeafCoin', datetime(2014, 1, 24)],
    '66':  ['66Coin', datetime(2014, 1, 26)],
    'tea': ['Teacoin', datetime(2014, 1, 27)],
    'huc': ['Huntercoin', datetime(2014, 1, 27)],
    'top': ['Topcoin', datetime(2014, 1, 27)],
    'max': ['Maxcoin', datetime(2014, 1, 29)],
    'utc': ['Ultracoin', datetime(2014, 2, 1)],
    'ben': ['Benjamins', datetime(2014, 2, 1)],
    'aur': ['Auroracoin', datetime(2014, 2, 2)],
    'pop': ['PopularCoin', datetime(2014, 2, 2)],
    'lbt': ['Litebar', datetime(2014, 2, 2)],
    'corg':['CorgiCoin', datetime(2014, 2, 2)],
    'karm':['Karmacoin', datetime(2014, 2, 4)],
    'hex': ['Heisenberg', datetime(2014, 2, 5)],
    'wtc': ['World Top Coin', datetime(2014, 2, 5)],
    'mint':['Mintcoin', datetime(2014, 2, 6)],
    'trl': ['TrollCoin', datetime(2014, 2, 7)],
    'sochi':['SochiCoin', datetime(2014, 2, 7)],
    'tak': ['TakeiCoin', datetime(2014, 2, 8)],
    'lyc': ['LycanCoin', datetime(2014, 2, 8)],
    'snp': ['SnapCoin', datetime(2014, 2, 9)],
    'lmc': ['LemonCoin', datetime(2014, 2, 9)],
    'akc': ['Anti-Keiser-Coin', datetime(2014, 2, 10)],
    'bee': ['BeeCoin', datetime(2014, 2, 12)],
    'epc': ['EmperorCoin', datetime(2014, 2, 13)],
    'doug':['DougCoin', datetime(2014, 2, 13)],
    'mim': ['Magic Internet Money', datetime(2014, 2, 13)],
    'loot':['Crypto LOOT', datetime(2014, 2, 13)],
    'flap':['Flappycoin', datetime(2014, 2, 14)],
    'note':['DNotes', datetime(2014, 2, 16)],
    'che': ['RevolutionCoin', datetime(2014, 2, 16)],
    'eqb': ['EquestrianBit', datetime(2014, 2, 17)],
    'v':   ['Version', datetime(2014, 2, 21)],
    'btq': ['BitQuark', datetime(2014, 2, 22)],
    'app': ['AppCoin', datetime(2014, 2, 23)],
    'bc':  ['BlackCoin', datetime(2014, 2, 24)],
    'mzc': ['MazaCoin', datetime(2014, 2, 24)],
    'blc': ['Blakecoin', datetime(2014, 2, 24)],
    'edu': ['EduCoin', datetime(2014, 2, 25)],
    'crd': ['Credits', datetime(2014, 2, 25)],
    'exe': ['EXECoin', datetime(2014, 2, 27)],
    'ccn': ['C-Coin', datetime(2014, 2, 27)],
    'zeit':['Zeitcoin', datetime(2014, 2, 28)],
    'uvc': ['UniversityCoin', datetime(2014, 2, 28)],
    'gox': ['MtgoxCoin', datetime(2014, 2, 28)],
    'emc2':['Einsteinium', datetime(2014, 3, 1)],
    'ddc': ['DokdoCoin', datetime(2014, 3, 1)],
    'flt': ['Fluttercoin', datetime(2014, 3, 5)],
    '888': ['OctoCoin', datetime(2014, 3, 8)],
    'fail':['FailCoin', datetime(2014, 3, 8)],
    'ghc': ['Ghostcoin', datetime(2014, 3, 10)],
    'blitz':['BlitzCoin', datetime(2014, 3, 10)],
    'diem':['Carpe Diem', datetime(2014, 3, 13)],
    'grs': ['GroestlCoin', datetime(2014, 3, 22)],
    'crypt':['CryptCoin', datetime(2014, 3, 25)],
    'comm':['Communitycoin', datetime(2014, 3, 25)],
    'sha': ['Sha1Coin', datetime(2014, 3, 28)],
    'wc':  ['Whitecoin', datetime(2014, 4, 4)],
    'adn': ['Aiden', datetime(2014, 4, 5)],
    'xlb': ['LibertyCoin', datetime(2014, 4, 5)],
    'ac':  ['AsiaCoin', datetime(2014, 4, 16)],
    'xmr': ['Monero', datetime(2014, 4, 18)],
    'cinni':['CinniCoin', datetime(2014, 4, 19)],
    'bst': ['GlobalBoost', datetime(2014, 4, 27)],
    'mon': ['Monocle', datetime(2014, 5, 2)],
    'cure':['Curecoin', datetime(2014, 5, 2)],
    'vrc': ['Vericoin', datetime(2014, 5, 10)],
    'nyc': ['NewYorkCoin', datetime(2014, 5, 21)],
    'shc': ['SherlockCoin', datetime(2014, 5, 28)],
    'cloak':['Cloakcoin', datetime(2014, 5, 31)],
    'trk': ['Truckcoin', datetime(2014, 6, 3)],
    'min': ['Minerals', datetime(2014, 6, 5)],
    'stl': ['STLCoin', datetime(2014, 6, 25)],
    'met': ['MetaCoin', datetime(2014, 7, 4)],
    'str': ['Stellar', datetime(2014, 7, 31)],
    'pxl': ['Pixelcoin', datetime(2014, 11, 1)],
    'pac': ['PacCoin', datetime(2014, 12, 12)],
    'dime':['Dimecoin', datetime(2014, 12, 23)],
}


# don't have launch dates for these...
#
# RubyCoin	RUBY	Date	POW Scrypt	60 Million Coins	M F E W	x
# SaturnCoin	SAT2	Date	POW Scrypt	7,777,777 Coins	M F E W	x
# Grumpycoin	GRUMP	Date	POW Scrypt	20 Billion Coins	M F E W	x
# XXLCOIN	XXL	Date	POW Scrypt	2800 Billion Coins	M F E W	x
# Stories	STY	Date	POW SHA256	60 Million Coins	M F E W	x
# PolishCoin	PCC	Date	POW Scrypt	150 Million Coins	M F E W	x
# AnimeCoin	ANI	Date	Quark Algo	1.976 Billion	M F E W	x
# BillionCoin	BIL	Date	Scrypt Algo	10 Billion	M F E W	x
# 365Coin	365	Date	SHA3 Keccak	365 Coins	M F E W	x
# PLNCoin	PLNC	Date	Scrypt Algo	38,540,000 Coins	M F E W	x
# SovereignCoin	SVC	Date	Type	Supply	M F E W	x
# 84Coin	84	Date	POW Scrypt	84 Coins	M F E W	x
# CrimeaCoin	CMA	Date	Type	100 Million	M F E W	x
# FckBanks	FCK	Date	POW Scrypt	84 Billion	M F E W	x
# BeliCoin	BELI	Date	POW/POS Scrypt	900 Billion	M F E W	x
# GPUCoin	GPUC	Date	POW Scrypt	13.5 Billion	M F E W	x
# RealStackCoin	RSC	Date	POW Scrypt Algo	134.4 Billion	M F E W	x
# W2Coin	W2C	2014	POW Scrypt	2.1 Billion Coins	M F E W	9300
# Riecoin', datetime(201402	POW prime constellations	84 Million coins	M F E W	x
# Maidsafecoin	SAFE	20060222	POR



from __future__ import print_function
import random
import requests

from concurrent import futures

useragent = 'moneywagon 1.4.3'

class ServiceDisagreement(Exception):
    pass

class NoService(Exception):
    pass

class SkipThisService(NoService):
    pass

class NoData(Exception):
    pass

class Service(object):
    """
    Represents a blockchain service running an Http interface.
    Some `Services` subclass will only support a subset of all pissible blockchain functions.
    All Services should subclass this class, and implement their own `get_*` method.
    """
    supported_cryptos = None # must be a list of lower case currency codes.

    def __init__(self, verbose=False, responses=None):
        self.responses = responses or {} # for caching
        self.verbose = verbose
        self.last_url = None
        self.last_raw_response = None

    def __repr__(self):
        return "<Service: %s (%s in cache)>" % (self.__class__.__name__, len(self.responses))

    def get_url(self, url, *args, **kwargs):
        return self._external_request('get', url, *args, **kwargs)

    def post_url(self, url, *args, **kwargs):
        return self._external_request('post', url, *args, **kwargs)

    def _external_request(self, method, url, *args, **kwargs):
        """
        Wrapper for requests.get with useragent automatically set.
        And also all requests are reponses are cached.
        """
        self.last_url = url
        if url in self.responses.keys() and method == 'get':
            return self.responses[url] # return from cache if its there

        headers = kwargs.pop('headers', None)
        custom = {'User-Agent': useragent}
        if headers:
            headers.update(custom)
            kwargs['headers'] = headers
        else:
            kwargs['headers'] = custom

        response = getattr(requests, method)(url, *args, **kwargs)

        if self.verbose:
            print("Got Response: %s" % url)

        if response.status_code == 503:
            raise SkipThisService("Service returned 503 - Temporarily out of service.")

        if method == 'get':
            self.responses[url] = response # cache for later

        self.last_raw_response = response
        return response

    def get_current_price(self, crypto, fiat):
        """
        Makes call to external service, and returns the price for given
        fiat/crypto pair. Returns two item tuple: (price, best_market)
        """
        raise NotImplementedError(
            "This service does not support getting the current fiat exchange rate. "
            "Or rather it has no defined 'get_current_price' method."
        )

    def get_historical_price(self, crypto, fiat, at_time):
        """
        """
        raise NotImplementedError(
            "This service does not support getting historical price."
            "Or rather it has no defined 'get_historical_price' method."
        )

    def get_transactions(self, crypto, address, confirmations=1):
        """
        Must be returned with the most recent transaction at the top.
        Returned is a list of dicts that have the following keys:

        `amount`: Number of units of currency moved. Always in base units (not satoshis).
        `date`: a datetime object of when the transaction was made.
        `txid`: The transaction ID, looks like a hash.
        `confirmations`: integer of the number of confirmations this transaction has on top of it.

        """
        raise NotImplementedError(
            "This service does not support getting historical transactions. "
            "Or rather it has no defined 'get_transactions' method."
        )

    def get_unspent_outputs(self, crypto, address):
        """
        Default implmentation of this function that uses get_transaction
        Subclasses should overwrite this with a direct call to get utxo (if applicable)

        Returned will be a list of dictionaries, the keys will be:

        required:

        `output` - the big endian tx hash, followed by a colon, then the tx index. (for pybitcointools support)
        `address` - the address passed in (for pybitcointools support)
        `amount` - int, satoshi amount of the input

        optional:

        `confirmations` - how many confirmations this tx has so far.

        """
        raise NotImplementedError(
            "This service does not support getting unspent outputs. "
            "Or rather it has no defined 'get_unspent_outputs' method."
        )

    def get_balance(self, crypto, address, confirmations=1):
        """
        Get the amount of coin in the address passed in.
        Always returns a single float.
        """
        raise NotImplementedError(
            "This service does not support getting address balances. "
            "Or rather it has no defined 'get_balance' method."
        )

    def push_tx(self, crypto, tx_hex):
        """
        Push transaction to the miner network. Returns txid if done
        successfully.
        """
        raise NotImplementedError(
            "This service does not support pushing transactions to the network. "
            "Or rather it has no defined 'push_tx' method."
        )

    def get_block(self, crypto, block_height='', block_number='', latest=False):
        """
        Get block based on either block height, block number or get the latest
        block. Only one of the previous arguments must be passed on.

        Returned is a dictionary object with the following keys:

        * required fields:

        block_number - int
        size - size of block
        time - datetime object of when the block was made
        hash - str (must be all lowercase)
        tx_count - int, the number of transactions included in thi block.

        * optional fields:

        confirmations - int
        sent_value - total value moved from all included transactions
        total_fees - total amount of tx fees from all included transactions
        mining_difficulty - what the difficulty was when this block was made.
        merkle_root - str (lower case)
        previous_hash - str (lower case)
        next_hash - str (lower case) (or `None` of its the latest block)
        """
        raise NotImplementedError(
            "This service does not support getting getting block data. "
            "Or rather it has no defined 'get_block' method."
        )

    def get_optimal_fee(self, crypto, tx_bytes, acceptable_block_delay):
        raise NotImplementedError(
            "This service does not support getting optimal fee. "
            "Or rather it has no defined 'get_optimal_fee' method."
        )


class AutoFallback(object):
    """
    Calls a succession of services until one returns a value.
    """

    def __init__(self, services=None, verbose=False, responses=None):
        """
        Each service class is instantiated here so the service instances stay
        in scope for the entire life of this object. This way the service
        objects can cache responses.
        """
        if not services:
            from moneywagon.services import ALL_SERVICES
            services = ALL_SERVICES

        self.services = []
        for ServiceClass in services:
            self.services.append(
                ServiceClass(verbose=verbose, responses=responses)
            )

        self.verbose = verbose

    def _try_each_service(self, method_name, *args, **kwargs):
        """
        Try each service until one returns a response. This function only
        catches the bare minimum of exceptions from the service class. We want
        exceptions to be raised so the service classes can be debugged and
        fixed quickly.
        """
        for service in self.services:
            crypto = ((args and args[0]) or kwargs['crypto']).lower()
            address = kwargs.get('address', '').lower()
            fiat = kwargs.get('fiat', '').lower()

            if service.supported_cryptos and (crypto not in service.supported_cryptos):
                if self.verbose:
                    print("SKIP:", "%s not supported for %s" % (crypto, service.__class__.__name__))
                continue
            try:
                if self.verbose: print("* Trying:", service, crypto, "%s%s" % (address, fiat))
                return getattr(service, method_name)(*args, **kwargs)
            except (KeyError, IndexError, TypeError, ValueError) as exc:
                # API has probably changed, therefore service class broken
                if self.verbose: print("FAIL:", exc.__class__.__name__, exc)
            except SkipThisService as exc:
                # service classes can raise this exception if for whatever reason
                # that service can't return a response, but maybe another one can.
                if self.verbose: print("SKIP:", exc.__class__.__name__, exc)
            except NotImplementedError as exc:
                if self.verbose: print("SKIP:", exc.__class__.__name__, exc)

        raise NoService(self.no_service_msg(*args, **kwargs))

    def no_service_msg(self, *args, **kwargs):
        """
        This function is called when all Services have been tried and no value
        can be returned. It much take the same args and kwargs as in the method
        spefified in `method_name`. Returned is a string for the error message.
        It should say something informative.
        """
        return "All either skipped or failed."

def enforce_service_mode(services, FetcherClass, kwargs, modes):
    """
    Fetches the value according to the mode of execution desired.
    `FetcherClass` must be a class that is subclassed from AutoFallback.
    `services` must be a list of Service classes.
    `kwargs` is a list of arguments used to make the service call, usually
      something like {crypto: 'btc', address: '1HwY...'} or
      {crypto: 'ltc', fiat: 'rur'}, (depends on the what FetcherClass.get takes)

    Modes can be:

         random = [True|False] False by default
         paranoid = positive int. 1 by default.
         fast = [True|False] False by default

    """
    paranoid_level = modes.get('paranoid', 1)
    verbose = modes.get('verbose', False)

    if modes.get('random', False):
        random.shuffle(services)

    if paranoid_level == 1:
        return FetcherClass(services=services, verbose=verbose).get(**kwargs)

    with futures.ThreadPoolExecutor(max_workers=len(services)) as executor:
        fetches = [
            executor.submit(
                FetcherClass(services=[service], verbose=verbose).get, **kwargs
            ) for service in services[:paranoid_level+2]
        ]

        results = []
        for future in futures.as_completed(fetches):
            results.append(future.result())

    if hasattr(FetcherClass, "strip_for_consensus"):
        to_compare = FetcherClass.strip_for_consensus(results)
    else:
        to_compare = results

    if len(set(to_compare)) == 1:
        # if all values match, return any one (in this case the first one).
        return results[0]
    else:
        raise ServiceDisagreement("Differing values returned: %s" % results)

def currency_to_protocol(amount):
    """
    Convert a string of 'currency units' to 'protocol units'. For instance
    converts 19.1 bitcoin to 1910000000 satoshis.

    Input is a float, output is an integer that is 1e8 times larger.

    It is hard to do this conversion because multiplying
    floats causes rounding nubers which will mess up the transactions creation
    process.

    examples:

    19.1 -> 1910000000
    0.001 -> 100000

    """
    if type(amount) == float:
        amount = "%.8f" % amount

    return int(amount.replace(".", '')) # avoiding float math



from __future__ import print_function

from concurrent import futures
from moneywagon import get_address_balance, get_current_price

def fetch_wallet_balances(wallets, fiat, **modes):
    """
    Wallets must be list of two item lists. First item is crypto, second item
    is the address. example:

    [
        ['btc', '1PZ3Ps9RvCmUW1s1rHE25FeR8vtKUrhEai'],
        ['ltc', 'Lb78JDGxMcih1gs3AirMeRW6jaG5V9hwFZ']
    ]
    """
    price_fetch = set([x[0] for x in wallets])
    balances = {}
    prices = {}

    fetch_length = len(wallets) + len(price_fetch)

    if not modes.get('async', False):
        # synchronous fetching
        for crypto in price_fetch:
            prices[crypto] = get_current_price(crypto, fiat, **modes)

        for crypto, address in wallets:
            balances[address] = get_address_balance(crypto, address.strip(), **modes)

    else:
        # asynchronous fetching
        if modes.get('verbose', False):
            print("Need to make", fetch_length, "external calls")

        with futures.ThreadPoolExecutor(max_workers=int(fetch_length / 2)) as executor:
            future_to_key = dict(
                (executor.submit(
                    get_current_price, crypto, fiat, **modes
                ), crypto) for crypto in price_fetch
            )

            future_to_key.update(dict(
                (executor.submit(
                    get_address_balance, crypto, address.strip(), **modes
                ), address) for crypto, address in wallets
            ))

            done, not_done = futures.wait(future_to_key, return_when=futures.FIRST_EXCEPTION)
            if len(done) > 0:
                #from ipdb import set_trace; set_trace()
                raise not_done.pop().exception()

            for future in done:
                key = future_to_key[future]
                if len(key) > 5: # this will break if a crypto symbol is longer than 5 chars.
                    which = balances
                else:
                    which = prices

                res = future.result()
                which[key] = res

    ret = []
    for crypto, address in wallets:
        crypto_value = balances[address]
        fiat_price, source = prices[crypto]
        ret.append({
            'crypto': crypto,
            'address': address,
            'crypto_value': crypto_value,
            'fiat_value': crypto_value * fiat_price,
            'conversion_price': fiat_price,
            'price_source': source
        })

    return ret



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

import sys

if sys.version_info <= (3,1):
    extra_install = ['futures',]
else:
    extra_install = []

setup(
    name="moneywagon",
    version='1.4.3',
    description='Next Generation Cryptocurrency Platform',
    long_description=open('README.md').read(),
    author='Chris Priest',
    author_email='cp368202@ohiou.edu',
    url='https://github.com/priestc/moneywagon',
    packages=find_packages(),
    scripts=['bin/moneywagon'],
    include_package_data=True,
    license='LICENSE',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
    ],
    install_requires=[
        'requests',
        'pytz',
        'arrow',
        'pybitcointools'
    ] + extra_install
)
